package cz.frantovo.nekurak.servlet;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Pattern;
import javax.servlet.ServletException;
import javax.servlet.ServletOutputStream;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

/**
 * <p>Servlet pro zpřístupnění fotek, které se nacházejí ve zvláštním adresáři.</p>
 *
 * <p>Má jeden inicializační parametr:</p>
 *  <ul>
 *     <li><code>adresar</code> – cesta k adresáři na disku, kde se nacházejí fotky.</li>
 *     <li>např. <code>/var/www/nekurak.net/fotky</code></li>
 *  </ul>
 *
 * <p>Očekává se struktura adresářů tohoto typu:</p>
 *  <ul>
 *	<li>Plné rozlišení: <code>/var/www/nekurak.net/fotky/original/1.jpg</code></li>
 *      <li>Náhled fotky:   <code>/var/www/nekurak.net/fotky/nahled/1.jpg</code></li>
 *  </ul>
 * <p>Lze ovlivnit konstantami níže.</p>
 *
 * <p>URL respektují fyzické umístění fotek na serveru (není např. ID v GET parametru stylem <code>/fotky?id=123</code>),
 * díky tomu je možné, aby fotky později servírovala přímo reverzní proxy a nemusely procházet přes naši aplikaci
 * (fotky jsou veřejné, není potřeba řešit autorizaci)</p>
 *
 * @author fiki
 */
public class Fotky extends HttpServlet {

    /** Název inicializačního parametru */
    private static final String INIT_ADRESAR = "adresar";
    /** Název podadresáře obsahujícího fotku v plném rozlišení */
    public static final String PODADRESAR_ORIGINAL = "original";
    /** Název podadresáře obsahujícího výchozí náhled fotky */
    public static final String PODADRESAR_NAHLED = "nahled";
    public static final String PRIPONA = "jpg";
    private static final String LOMITKO = File.separator;
    /** Regulární výraz */
    private static final String VZOR_CESTY = "^" + LOMITKO + "(" + PODADRESAR_ORIGINAL + "|" + PODADRESAR_NAHLED + ")" + LOMITKO + "\\d+\\." + PRIPONA + "$";
    private static final String MIME_TYP = "image/jpeg";
    private File adresar;
    private static final Logger log = Logger.getLogger(Fotky.class.getSimpleName());

    @Override
    public void init() throws ServletException {
	super.init();
	String initAdresar = getServletConfig().getInitParameter(INIT_ADRESAR);
	adresar = new File(initAdresar);
	if (adresar.isDirectory()) {
	    log.log(Level.INFO, "Servlet „Fotka“ byl úspěšně inicializován.");
	    log.log(Level.INFO, "Adresář s fotkami: " + initAdresar);
	    log.log(Level.INFO, "RegExp cesty: " + VZOR_CESTY);
	} else {
	    throw new ServletException("Servlet „Fotka“ se nepodařilo inicializovat. Cesta: " + initAdresar);
	}
    }

    /**
     * @param pozadavek pouze GET (není důvod podporovat POST)
     * @param odpoved odešleme fotku s MIME typem podle konstanty, délkou a datem podle souboru.
     * @throws ServletException pokud je požadovaná cesta chybná (nevyhovuje vzoru)
     * @throws IOException
     */
    @Override
    protected void doGet(HttpServletRequest pozadavek, HttpServletResponse odpoved) throws ServletException, IOException {

	String cesta = zkontrolujParametr(pozadavek.getPathInfo());
	File soubor = new File(adresar, cesta);

	if (soubor.isFile() && soubor.canRead()) {

	    if (soubor.lastModified() > pozadavek.getDateHeader("If-Modified-Since")) {
		/** Soubor se změnil nebo ho klient ještě nemá načtený. */
		odpoved.setContentType(MIME_TYP);
		odpoved.setContentLength((int) soubor.length());
		odpoved.setDateHeader("Last-Modified", soubor.lastModified());

		ServletOutputStream out = odpoved.getOutputStream();
		InputStream in = new FileInputStream(soubor);

		try {
		    byte[] bytes = new byte[1024];
		    int bajtuNacteno;
		    while ((bajtuNacteno = in.read(bytes)) != -1) {
			out.write(bytes, 0, bajtuNacteno);
		    }
		} catch (Exception e) {
		    throw new ServletException("Chyba při odesílání obrázku klientovi.", e);
		} finally {
		    in.close();
		    out.close();
		}
	    } else {
		/** Soubor se od posledního načtení klientem nezměnil → není potřeba ho posílat znova. */
		odpoved.setStatus(HttpServletResponse.SC_NOT_MODIFIED);
	    }

	} else {
	    /** Neexistující nebo nečitelný soubor → HTTP 404 chyba */
	    odpoved.sendError(HttpServletResponse.SC_NOT_FOUND);
	}
    }

    /**
     * @param cesta cesta požadovaná klientem: <code>request.getPathInfo()</code>
     * @throws ServletException pokud cesta nevyhovuje vzoru
     */
    private static String zkontrolujParametr(String cesta) throws ServletException {
	if (Pattern.matches(VZOR_CESTY, cesta)) {
	    /** cesta je v pořádku → pokračujeme */
	    return cesta;
	} else {
	    /** Chybná cesta → HTTP 500 chyba */
	    throw new ServletException("Chybná cesta k obrázku: " + cesta);
	}
    }
}
